/*
 * kmscon - Fixed 8x16 font for font handling of text renderer
 *
 * Copyright (c) 2012 David Herrmann <dh.herrmann@googlemail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * SECTION:text_font_8x16.c
 * @short_description: Fixed 8x16 font for font handling of text renderer
 * @include: text.h
 *
 * This is a fixed font renderer backend that supports just one font which is
 * statically compiled into the file. This font is a very simple 8x16 font with
 * several special chars according to DEC-special-sets and common linux kernel
 * character-sets.
 *
 * When loading a font-face via this backend, then the static font is always
 * returned. This means, we need no internal state and can instead share the
 * buffer without locking. Every character outside of Latin1 is ignored so most
 * Unicode characters cannot be drawn with this backend.
 *
 * There is also no sophisticated font handling in here so this should only be
 * used as last fallback when external libraries failed. However, this also
 * means this backend is _very_ fast as no rendering is needed. Everything is
 * pre-rendered. See the big array at the end of this file for the bitmap.
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "log.h"
#include "text.h"
#include "unicode.h"
#include "uterm.h"

#define LOG_SUBSYSTEM "text_font_8x16"

/* forward declaration; see end of file for real array */
static const struct kmscon_glyph kmscon_font_8x16_glyphs[256];

static int kmscon_font_8x16_init(struct kmscon_font *out,
				 const struct kmscon_font_attr *attr)
{
	static const char name[] = "static-8x16";

	log_debug("loading static 8x16 font");

	memset(&out->attr, 0, sizeof(out->attr));
	memcpy(out->attr.name, name, sizeof(name));
	out->attr.bold = false;
	out->attr.italic = false;
	out->attr.width = 8;
	out->attr.height = 16;
	kmscon_font_attr_normalize(&out->attr);

	return 0;
}

static void kmscon_font_8x16_destroy(struct kmscon_font *font)
{
	log_debug("unloading static 8x16 font");
}

static int kmscon_font_8x16_render(struct kmscon_font *font,
				   kmscon_symbol_t sym,
				   const struct kmscon_glyph **out)
{
	const uint32_t *val;
	size_t len;

	val = kmscon_symbol_get(&sym, &len);
	if (len > 1 || *val >= 256)
		return -ERANGE;

	*out = &kmscon_font_8x16_glyphs[*val];

	return 0;
}

static void kmscon_font_8x16_drop(struct kmscon_font *font,
				  const struct kmscon_glyph *glyph)
{
}

static const struct kmscon_font_ops kmscon_font_8x16_ops = {
	.name = "8x16",
	.init = kmscon_font_8x16_init,
	.destroy = kmscon_font_8x16_destroy,
	.render = kmscon_font_8x16_render,
	.drop = kmscon_font_8x16_drop,
};

int kmscon_font_8x16_load(void)
{
	int ret;

	ret = kmscon_font_register(&kmscon_font_8x16_ops);
	if (ret) {
		log_error("cannot register 8x16 font");
		return ret;
	}

	return 0;
}

void kmscon_font_8x16_unload(void)
{
	kmscon_font_unregister(kmscon_font_8x16_ops.name);
}

static const struct kmscon_glyph kmscon_font_8x16_glyphs[256] = {
	{ /* 0 0x00 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 1 0x01 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 2 0x02 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 3 0x03 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 4 0x04 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 5 0x05 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 6 0x06 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 7 0x07 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 8 0x08 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
			},
		},
	},
	{ /* 9 0x09 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 10 0x0a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
			},
		},
	},
	{ /* 11 0x0b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 12 0x0c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 13 0x0d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 14 0x0e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 15 0x0f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 16 0x10 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 17 0x11 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 18 0x12 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 19 0x13 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 20 0x14 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 21 0x15 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 22 0x16 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 23 0x17 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 24 0x18 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 25 0x19 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 26 0x1a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 27 0x1b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 28 0x1c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 29 0x1d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 30 0x1e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 31 0x1f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 32 0x20 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 33 0x21 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 34 0x22 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 35 0x23 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 36 0x24 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 37 0x25 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 38 0x26 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 39 0x27 '*/
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 40 0x28 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 41 0x29 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 42 0x2a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 43 0x2b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 44 0x2c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 45 0x2d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 46 0x2e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 47 0x2f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 48 0x30 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 49 0x31 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 50 0x32 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 51 0x33 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 52 0x34 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 53 0x35 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 54 0x36 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 55 0x37 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 56 0x38 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 57 0x39 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 58 0x3a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 59 0x3b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 60 0x3c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 61 0x3d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 62 0x3e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 63 0x3f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 64 0x40 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 65 0x41 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 66 0x42 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 67 0x43 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 68 0x44 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 69 0x45 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 70 0x46 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 71 0x47 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 72 0x48 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 73 0x49 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 74 0x4a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 75 0x4b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 76 0x4c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 77 0x4d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 78 0x4e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 79 0x4f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 80 0x50 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 81 0x51 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 82 0x52 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 83 0x53 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 84 0x54 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 85 0x55 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 86 0x56 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 87 0x57 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 88 0x58 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 89 0x59 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 90 0x5a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 91 0x5b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 92 0x5c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 93 0x5d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 94 0x5e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 95 0x5f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 96 0x60 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 97 0x61 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 98 0x62 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 99 0x63 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 100 0x64 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 101 0x65 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 102 0x66 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 103 0x67 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 104 0x68 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 105 0x69 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 106 0x6a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 107 0x6b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 108 0x6c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 109 0x6d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 110 0x6e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 111 0x6f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 112 0x70 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 113 0x71 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 114 0x72 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 115 0x73 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 116 0x74 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 117 0x75 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 118 0x76 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 119 0x77 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 120 0x78 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 121 0x79 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 122 0x7a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 123 0x7b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 124 0x7c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 125 0x7d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 126 0x7e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 127 0x7f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 128 0x80 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 129 0x81 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 130 0x82 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 131 0x83 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 132 0x84 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 133 0x85 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 134 0x86 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 135 0x87 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 136 0x88 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 137 0x89 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 138 0x8a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 139 0x8b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 140 0x8c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 141 0x8d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 142 0x8e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 143 0x8f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 144 0x90 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 145 0x91 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 146 0x92 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 147 0x93 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 148 0x94 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 149 0x95 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 150 0x96 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 151 0x97 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 152 0x98 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 153 0x99 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 154 0x9a */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 155 0x9b */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 156 0x9c */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 157 0x9d */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 158 0x9e */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 159 0x9f */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 160 0xa0 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 161 0xa1 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 162 0xa2 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 163 0xa3 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 164 0xa4 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 165 0xa5 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 166 0xa6 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 167 0xa7 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 168 0xa8 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 169 0xa9 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 170 0xaa */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 171 0xab */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 172 0xac */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 173 0xad */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 174 0xae */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 175 0xaf */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 176 0xb0 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0x00, 0xff, 0x00, 0x00,
				0x00, 0xff, 0x00, 0x00,
			},
		},
	},
	{ /* 177 0xb1 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
				0x00, 0xff, 0x00, 0xff,
				0x00, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0x00,
				0xff, 0x00, 0xff, 0x00,
			},
		},
	},
	{ /* 178 0xb2 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
			},
		},
	},
	{ /* 179 0xb3 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 180 0xb4 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 181 0xb5 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 182 0xb6 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 183 0xb7 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 184 0xb8 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 185 0xb9 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 186 0xba */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 187 0xbb */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 188 0xbc */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 189 0xbd */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 190 0xbe */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 191 0xbf */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 192 0xc0 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 193 0xc1 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 194 0xc2 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 195 0xc3 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 196 0xc4 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 197 0xc5 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 198 0xc6 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 199 0xc7 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 200 0xc8 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 201 0xc9 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 202 0xca */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 203 0xcb */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 204 0xcc */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 205 0xcd */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 206 0xce */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 207 0xcf */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 208 0xd0 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 209 0xd1 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 210 0xd2 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 211 0xd3 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 212 0xd4 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 213 0xd5 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 214 0xd6 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 215 0xd7 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
			},
		},
	},
	{ /* 216 0xd8 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 217 0xd9 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 218 0xda */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 219 0xdb */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
			},
		},
	},
	{ /* 220 0xdc */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
			},
		},
	},
	{ /* 221 0xdd */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 222 0xde */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
			},
		},
	},
	{ /* 223 0xdf */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 224 0xe0 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 225 0xe1 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 226 0xe2 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 227 0xe3 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 228 0xe4 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 229 0xe5 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 230 0xe6 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 231 0xe7 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 232 0xe8 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 233 0xe9 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 234 0xea */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 235 0xeb */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 236 0xec */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 237 0xed */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 238 0xee */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 239 0xef */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 240 0xf0 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 241 0xf1 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 242 0xf2 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 243 0xf3 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 244 0xf4 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0xff, 0xff,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 245 0xf5 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 246 0xf6 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 247 0xf7 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 248 0xf8 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 249 0xf9 */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 250 0xfa */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 251 0xfb */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0xff, 0xff,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0xff, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 252 0xfc */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 253 0xfd */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0xff, 0xff, 0x00, 0x00,
				0x00, 0x00, 0x00, 0xff,
				0xff, 0x00, 0x00, 0x00,
				0x00, 0x00, 0xff, 0xff,
				0x00, 0x00, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 254 0xfe */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0xff, 0xff, 0xff,
				0xff, 0xff, 0xff, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
	{ /* 255 0xff */
		.ascent = 12,
		.descent = 4,
		.buf = {
			.width = 8,
			.height = 16,
			.stride = 8,
			.format = UTERM_FORMAT_GREY,
			.data = (uint8_t[]){
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
				0x00, 0x00, 0x00, 0x00,
			},
		},
	},
};
